from ete2 import PhyloTree


def load_tree_sequences(nwk_file, fasta_file):
    ''' Load a tree with associated sequences on leaves. '''
    tree = PhyloTree(newick = nwk_file, format = 1)
    tree.link_to_alignment(alignment = fasta_file, alg_format = 'fasta')
    return tree


def filter_leaves(tree, node_name_list):
    if not len(node_name_list):
        return tree.get_leaves() # no filtering -- all leaves returned
    return filter(lambda node: node.name in node_name_list, tree.get_leaves())
    
def get_mrca(tree, nodes):
    return tree.get_common_ancestor(nodes)

def get_mrca_leaf_names(tree, node_name_list):
    node_subset = filter_leaves(tree, node_name_list)
    return get_mrca(tree, node_subset)

def get_mrca_subtree(tree_nwk_file, fasta_file, sel_annot_nodes):
    ''' 
    Get subtree whose leaves all have the same annotation
    Typically finds subtree where all sequences evolve same 
    function
    '''
    tree = load_tree_sequences(tree_nwk_file, fasta_file)

    # Get list of tree leaf objects with selected function
    sel_annot_leaves = filter_leaves(tree, sel_annot_nodes)

    # Get common ancestor for leaves with function of interest
    anc_node = get_mrca(tree, sel_annot_leaves)
    return anc_node, sel_annot_leaves




#! /usr/bin/env python 
import sys
import re
from collections import defaultdict

# List manipulation
flatten = lambda x: [subitem for item in x for subitem in item]
list_pairs = lambda x: [(x[i1],x[i2]) for i1 in range(len(x)-1) \
                       for i2 in range(i1+1,len(x))]
#integer substring
get_int = lambda s: int(re.sub('\D','',s))

#lists of strings
sort_by_digits = lambda x: sorted(x, key = lambda s: get_int(s))

#list sorting
sort_list_by_other = lambda x,y: [x for (y,x) in sorted(zip(y,x),
                                        key=lambda pair: pair[0])]


# General IO
def stderr_write(myList):
    sys.stderr.write(' '.join([str(el) for el in myList])+'\n')


def parse_column(infile, col_num = 0):
    out_list = []
    with open(infile, 'r') as f:
        for line in f:
            recs = line.replace('"','').strip().split()
            if len(recs) > col_num:
                out_list.append(recs[col_num])
    return out_list

def parse_keyval_dict(infile):
    ''' 
    Simple {key:val} dict from first two columns 
    in file 
    '''
    out_dict = {}
    with open(infile, 'rb') as f:
        for line in f:
            recs = line.replace('"','').strip().split()
            if len(recs) > 1:
                out_dict[recs[0]] = recs[1]
    return out_dict

def parse_keyval_dlist(infile):
    ''' 
    List-valued {key:[val0,val1,...]} dict from 
    first two columns in file 
    '''
    out_dict = defaultdict(list)
    with open(infile, 'rb') as f:
        for line in f:
            recs = line.replace('"','').strip().split()
            if len(recs) > 1:
                out_dict[recs[0]].append(recs[1])
    return out_dict




#! /usr/bin/env python 

# Notes: consider removing ete2/3 dependence
# ete is GPL + only need the PhyloTree class implementation
# https://github.com/etetoolkit/ete/blob/master/ete3/phylo/phylotree.py
# need to be able to read Newick (*.nwk) files and build a 
# general (nonbinary) phylo tree with attributes on nodes and edges 
# FORMAT  DESCRIPTION
# 1       flexible with internal node names


from scipy import stats
from ete2 import PhyloTree

from napa.utils.general import *
from napa.phylo.tree.edge import *
from napa.phylo.tree.tree import *

class PhyloMut(object):
    '''
    Similar to AlnMut class, but counts phylo edges instead of 
    alignmen sequences of occurrence.
    '''
    def __init__(self, phylo_edge_pairs = set(), mut_str = None):
        self.mut_str = mut_str
        self.phylo_edge_pairs = set(phylo_edge_pairs)
    
    def add_phylo_edge_pair(self,phylo_edge_pair):
        self.phylo_edge_pairs.add(phylo_edge_pair) 
    
    def add_phylo_edge_pairs(self,phylo_edge_pairs):
        self.phylo_edge_pairs.update(phylo_edge_pairs)

    def __repr__(self):
        return self.mut_str


class PhyloMutPair(object):
    '''
    Pair of muts from tree.
    '''    
    def __init__(self, phylo_edge_pairs = set(), mut_pair = None, 
                 all_edges = set(), prot_func_transitions = {'':''}):

        # List of pairs of phylo_edge_objects
        if not len(phylo_edge_pairs):
            raise ValueError('In PhyloMutPair: ' + \
                             'Set of Phylo edge pairs must be nonempty!')
        else:
            self.phylo_edge_pairs = set(phylo_edge_pairs)

        # A unique pair of AA muts
        if mut_pair ==  None or not len(mut_pair):
            raise ValueError('In PhyloMutPair: No mutations given!')
        else:
            self.mut_pair = mut_pair
            self.name = '%s_%s' % (mut_pair[0].mut_str, mut_pair[1].mut_str)
        if not len(all_edges):
            raise ValueError('In PhyloMutPair: No edge pairs having each ' + \
                             'mutation were given!')
        else:
            self.all_edges = set(all_edges)
        self.prot_func_transitions = prot_func_transitions

    def add_phylo_edge_pair(self, phylo_edge_pair):
        self.phylo_edge_pairs.add(phylo_edge_pair) 
    
    def add_phylo_edge_pairs(self, phylo_edge_pairs):
        self.phylo_edge_pairs.update(phylo_edge_pairs)


    def assign_lineage_distribution(self):
        common_lineage, mut0_lineage, mut1_lineage = [], [], []
        for pep in self.phylo_edge_pairs:
            if pep[0] == pep[1]:
                common_lineage  +=  [pep[0].parent] + [pep[0]] + pep[0].followers 
            else:
                common_prec = pep[0].get_common_precs(pep[1], 'function', 
                                                      self.prot_func_transitions)
                common_follow = list(set([pep[0]] + pep[0].followers) & \
                                    set([pep[1]] + pep[1].followers))
                common_lineage  +=  common_prec + common_follow

                mut0_prec = pep[0].get_intermediate_prec(common_prec, 'function',
                                                         self.prot_func_transitions)
                mut1_prec = pep[1].get_intermediate_prec(common_prec, 'function',
                                                         self.prot_func_transitions)

                mut0_lineage += list(set(mut0_prec + pep[0].followers) - \
                                     set(common_prec + common_follow))
                mut1_lineage += list(set(mut1_prec + pep[1].followers) - \
                                     set(common_prec + common_follow))
        
        self.common_lineage = set([p for p in common_lineage if p.muts])  
        self.mut0_lineage = set([p for p in mut0_lineage if p.muts]) - \
                            self.common_lineage
        self.mut1_lineage = set([p for p in mut1_lineage if p.muts]) - \
                            self.common_lineage
        assoc_lineage = self.common_lineage | self.mut0_lineage | self.mut1_lineage
        self.other_lineage = self.all_edges - assoc_lineage
        #self.print_lineages()


    def print_lineages(self):
        stderr_write(["\tCommon Lineage:", 
                      [p.name for p in self.common_lineage]])
        stderr_write(["\tFirst mutation only lineage:", 
                      [p.name for p in self.mut0_lineage]])
        stderr_write(["\tSecond mutations only lineage:", 
                     [p.name for p in self.mut1_lineage]])
        stderr_write(["\tOther mutations lineage:", 
                     [p.name for p in self.other_lineage]])


    def get_contingency_table(self):
        '''
        Contingency table for Fisher's Exact Test.
        Uses overall size of clades where mutations in pair
        occurred. 
        '''
        common = len(self.common_lineage)
        only0 = len(self.mut0_lineage) 
        only1 = len(self.mut1_lineage)
        other = len(self.other_lineage)
        self.contingency_table = [[common, only0], [only1, other]]


    def get_contingency_table_edge_counts(self):
        ''' 
        A simpler contingency table based on counts of co-occurrence of 
        muts in tree edges rather than clades. 
        Looking at clade size tends to be more informative.
        '''
        occur01 = len(self.phylo_edge_pairs)
        occur0 = len(self.mut_pair[0].phylo_edge_pairs - self.phylo_edge_pairs) 
        occur1 = len(self.mut_pair[1].phylo_edge_pairs - self.phylo_edge_pairs)
        occur_other = len(self.all_phylo_edge_pairs\
                        -self.mut_pair[0].phylo_edge_pairs \
                        -self.mut_pair[1].phylo_edge_pairs)
        self.contingency_table = [[occur01, occur0], [occur1, occur_other]]
        

    def get_fisher_pval(self):
        '''
        Uses scipy.stats to calculate Fisher's exact p-values based 
        on the contingency table. For association, only need the 
        right-tailed p-value: pval_more.
        '''
        self.assign_lineage_distribution()
        self.get_contingency_table()
        self.odds_ratio_less, self.p_val_less = \
                        stats.fisher_exact(self.contingency_table, 
                                           alternative = 'less')

        self.odds_ratio_more, self.p_val_more = \
                        stats.fisher_exact(self.contingency_table, 
                                           alternative = 'greater')

class PhyloMutPairSet(object):

    def __init__(self, leaf_fasta_file = '', 
                 int_node_fasta_file = '', aln_pos = [],
                 wt_seq_str = '', wt_id = '', pos_subset = [], 
                 leaf_seqid_to_prot_func = {'':''}, 
                 leaf_seqid_to_prot_func_file = '', 
                 int_seqid_to_prot_func = {'':''}, 
                 int_seqid_to_prot_func_file = '', 
                 prot_func_transitions = {},
                 prot_func_transitions_file = '',
                 sel_prot_func_list = [''],
                 sel_prot_func_file = '',
                 dist_thresh = 0.,
                 tree_nwk_file = '',
                 mut_pair_type = 'dir'):
        
        # Get sequences for ALL nodes on tree
        # Sequences from MSA used in phylogeny reconstruction
        # (These end up on tree leaves) AND
        # Reconstructed internal node sequences
        self.aln = BioSeqAln(aln_fasta_file = leaf_fasta_file,
                             aln_pos = aln_pos, 
                             annot_key = 'function', 
                             seqid_to_annot = leaf_seqid_to_prot_func,
                             seqid_to_annot_file = leaf_seqid_to_prot_func_file)
        # Add reconstructed internal node sequences
        self.aln.add_sequences_from_file(int_node_fasta_file)

        # Option to annotate sequences by (reconstructed) function
        # Add reconstructed internal node sequences functions
        self.aln.annot_seqs(annot_key = 'function',
                            seqid_to_annot_file = int_seqid_to_prot_func_file,
                            seqid_to_annot = int_seqid_to_prot_func)

        # Option to focus on protein region(s)/domain(s) of interest
        self.aln.subset_aln_pos(pos_subset = pos_subset)

        # Get mutations from ancestral/wt sequence 
        # for each sequence in alignment
        # option to focus on subset of aln./protein positions
        self.get_wt_seq(wt_id, wt_seq_str)
        self.wt_seq.extract_pos(pos_subset)

        # Python dict representing selected transiotions
        # between reconstructed functions in tree/alignment
        # If empty (default) all functions considered
        # Example, when considering evolution of 
        # extended-spectrum resistance (2be)
        # {'2b':'2be', '2be':'2be', '2br':'2be', '2br':'2ber'}
        self.get_prot_func_transitions(prot_func_transitions, 
                                       prot_func_transitions_file)

        # Names/ids of all tree nodes having selected function
        self.get_sel_func_nodes(annot_key = 'function',
                                annot_list_file = sel_prot_func_file,
                                annot_list = sel_prot_func_list)

        # Get the sequence ids of all leaf sequences annotated
        # with the selected function and find most recent common ancestor
        # This isolates subtree with given function
        self.anc_node, self.sel_func_leaves = \
                    get_mrca_subtree(tree_nwk_file = tree_nwk_file,
                                     fasta_file = leaf_fasta_file,
                                     sel_annot_nodes = self.sel_func_node_ids)


        # Get muts for seqs corresponding to protein function
        self.get_leaf_func_muts()

        # Set of all phylogenetic tree edges
        #self.all_phylo_edge_pairs = set()

        # The folling generates a linegraph object of the whole tree
        # it is a tree rooted in the edge between anc_node and its ancestor
        # nodes of PhyloEdge are edges are in the original tree
        self.anc_edge = PhyloEdge(parent_node = self.anc_node.up, 
                                  child_node = self.anc_node,
                                  aln = self.aln, wt_seq = self.wt_seq)

        self.anc_edge.build_edge_tree()

        # Get set of all pairs of mutations occurring along tree
        # default directed mutation pairs
        self.mut_pair_type = 'undir' if 'undir' in mut_pair_type.lower() \
                             else 'dir'
        self.get_mut_pairs(dist_thresh, self.mut_pair_type)
        self.mut_pairs_fisher_pvalues()

    def get_wt_seq(self, wt_id, wt_seq_str):
        '''
        Obtain WT/ancestral sequence to which all other sequences in 
        the alignment/phylogeny should be compared. Can provide id+sequence,
        or the id of a sequence already in the alignment.
        '''
        if len(wt_id) and len(wt_seq_str) == self.aln.length:
            self.wt_seq = BioSeq(seq_id = wt_id, seq_str = wt_seq_str,
                                 seq_type = 'Protein', 
                                 seq_pos_list = self.aln_pos)
        elif len(wt_id) and len(wt_seq_str) != self.aln.length:
            if wt_id in self.aln.seqid_to_seq:
                #stderr_write(['AlnMutPairSet WARNING:',
                #              'Sequence length does not match alignment,',
                #              'but sequence id in alignment. Replacing',
                #              'with corresp. alignment sequence.'])
                self.wt_seq = self.aln.seqid_to_seq[wt_id]
            else:
                stderr_write(['Invalid  input for WT sequence.',
                              'Please enter valid id already in alignment.',
                               'or a new sequence.'])
                              
        elif not len(wt_id) and len(wt_seq_str) == self.aln.length:
            self.wt_seq = BioSeq(seq_id = 'Wild_type', 
                                 seq_str = wt_seq_str,
                                 seq_type = 'Protein', 
                                 seq_pos_list = self.aln_pos)

        else:
            stderr_write(['Invalid  input for WT sequence.',
                          'Please enter valid id already in alignment.',
                          'or a new sequence.'])


    def get_sel_func_nodes(self, annot_key = 'function',
                           annot_list_file = '',
                           annot_list = ['']):
        if annot_list == [''] and len(annot_list_file):
            annot_list = parse_column(annot_list_file)
        
        self.sel_func_node_ids = self.aln.seqids_with_annot(\
                                    annot_key = annot_key,
                                    annot_list = annot_list)
        
    def get_leaf_func_muts(self):
        '''
        Get muts associated with a function that occur in 
        selected tree leaves.
        This is the subset of all mutations on the tree that is most 
        associated with the function of interest.
        '''
        self.func_muts = []
        for seqid in [leaf.name for leaf in self.sel_func_leaves]:
            self.func_muts += self.wt_seq.get_substitutions(\
                                        self.aln.seqid_to_seq[seqid])
        self.func_muts = set(self.func_muts)

    def get_prot_func_transitions(self, prot_func_transitions = {}, 
                                  prot_func_transitions_file = ''):
        if not len(prot_func_transitions):
            self.prot_func_transitions = \
                        parse_keyval_dlist(prot_func_transitions_file)
        else:
            self.prot_func_transitions = prot_func_transitions

    def check_edge_pair(self,start_edge, follow_edge, dist_thresh):
        if start_edge.is_leaf_edge(): 
            return False
        if not self.check_edge(start_edge, dist_thresh):
            return False
        if not self.check_edge(follow_edge, dist_thresh):
            return False
        if not start_edge.check_dist(follow_edge, dist_thresh):
            return False
        return True

    def check_edge(self, edge, dist_thresh):
        if not edge.check_node_seq_annot('function', 
                                         self.prot_func_transitions):
            return False

        if not len(set(edge.muts) & self.func_muts):
            return False
        # Edge is too long -- exceeds distance thresholds
        if not edge.check_dist(edge, dist_thresh):
            return False
            
        return True
        

    def add_phylo_edge_pair(self, phylo_edge_pair, mut_list_pair):
        '''
        Add a phylogeny edge pair on which edges each of the 
        two mutations occurs.
        mut_list_pair: mutation lists for first and second edge
        in phylo_edge_pair
        Also keep track of individual mutation occurrence.
        '''
        mut_set = set(mut_list_pair[0]+ mut_list_pair[1])
        for mut_str in mut_set:
            if mut_str not in self.mut_str_to_obj:
                self.mut_str_to_obj[mut_str] = \
                    PhyloMut(phylo_edge_pairs = set(phylo_edge_pair),
                             mut_str = mut_str)
            else:
                self.mut_str_to_obj[mut_str].add_phylo_edge_pair(phylo_edge_pair)

        for mut_str_pair in zip(mut_list_pair[0], mut_list_pair[1]):
            if mut_str_pair not in self.mut_str_pair_to_obj:
                self.mut_str_pair_to_obj[mut_str_pair] = \
                    PhyloMutPair(phylo_edge_pairs = set([phylo_edge_pair]),
                                 mut_pair = [self.mut_str_to_obj[mut_str_pair[0]],
                                             self.mut_str_to_obj[mut_str_pair[1]]],
                                 all_edges = self.all_edges)
            else:
                self.mut_str_pair_to_obj[mut_str_pair].add_phylo_edge_pair(\
                                                                    phylo_edge_pair)


    def get_mut_pairs(self, dist_thresh, mut_pair_type):
        ''' 
        Extract pairs of mutations occurring along tree edges
        '''
        self.all_edges = [self.anc_edge] + \
                         [fe for fe in self.anc_edge.followers]
        self.all_edges = [e for e in self.all_edges \
                          if self.check_edge(e, dist_thresh)]
        
        stderr_write(['Number of tree edges considered:', 
                      len(self.all_edges)])

        # will hold sets of mutation(s)/ (pairs)
        self.mut_str_pair_to_obj, self.mut_str_to_obj = {}, {}

        for start_edge in self.all_edges:
            start_edge_sorted_muts = sort_by_digits(list(\
                                                    set(start_edge.muts) & \
                                                         self.func_muts))
            if not len(start_edge_sorted_muts): continue
            # Make sure mutation pairs occurring on same edge included
            if len(start_edge_sorted_muts) > 1:
                self.add_phylo_edge_pair(phylo_edge_pair = (start_edge, start_edge), 
                                         mut_list_pair = \
                                         zip(*list_pairs(start_edge_sorted_muts)))
            
            if 'undir' in mut_pair_type.lower():
                self.add_undir_pairs(start_edge, start_edge_sorted_muts, 
                                     dist_thresh)
            else:
                self.add_dir_pairs(start_edge, start_edge_sorted_muts, 
                                   dist_thresh)
                
        stderr_write(['Number of mutations associated with function:', 
                      len(self.mut_str_to_obj.keys())])
        stderr_write(['Number of mutation pairs extracted', 
                      len(self.mut_str_pair_to_obj.keys())])


    def add_undir_pairs(self, start_edge, start_edge_sorted_muts, dist_thresh):
        for follow_edge in start_edge.followers:
            follow_edge_sorted_muts = sorted(list((set(follow_edge.muts) & \
                                         self.func_muts) - set(start_edge.muts)))

            if self.check_edge_pair(start_edge, follow_edge, dist_thresh):
                # For undirected mutation pairs, mutations are ordered by 
                # residue position
                mut_pairs = [(sm, fm) for sm in start_edge_sorted_muts \
                             for fm in follow_edge_sorted_muts \
                             if get_int(sm) <= get_int(fm)]
                self.add_phylo_edge_pair(phylo_edge_pair = (start_edge, follow_edge), 
                                        mut_list_pair = [start_edge_sorted_muts,
                                                         follow_edge_sorted_muts])

            # Reverse order of mutations in cases where follower position
            # is less than start position and distance from follower to start
            # is less than threshold
            if self.check_edge_pair(follow_edge, start_edge, dist_thresh):
                rev_mut_pairs = [(fm, sm) for sm in start_edge_sorted_muts \
                                 for fm in follow_edge_sorted_muts \
                                 if get_int(sm) > get_int(fm)]
                self.add_phylo_edge_pair(phylo_edge_pair = (follow_edge, start_edge), 
                                         mut_list_pair = [follow_edge_sorted_muts,
                                                         start_edge_sorted_muts])


    def add_dir_pairs(self, start_edge, start_edge_sorted_muts, dist_thresh):
        for follow_edge in start_edge.followers:
            follow_edge_sorted_muts = sorted(list((set(follow_edge.muts) & \
                                         self.func_muts) - set(start_edge.muts)))

            if self.check_edge_pair(start_edge, follow_edge, dist_thresh):
                # For undirected mutation pairs, mutations are ordered by 
                # residue position
                mut_pairs = [(sm, fm) for sm in start_edge_sorted_muts \
                             for fm in follow_edge_sorted_muts]
                self.add_phylo_edge_pair(phylo_edge_pair = (start_edge, follow_edge), 
                                         mut_list_pair = [start_edge_sorted_muts,
                                                          follow_edge_sorted_muts])


    def mut_pairs_fisher_pvalues(self):
        '''
        Calculate Fisher's Exact Test p-value, 
        for each PhyloMutPair object
        '''
        for mut_str_pair in self.mut_str_pair_to_obj:
            mut_pair = self.mut_str_pair_to_obj[mut_str_pair]
            mut_pair.get_fisher_pval()


    def __repr__(self):
        out_str = '\t'.join(['_Mut1_', '_Mut2_', 'edgePairList', 
                          'contTable(common_1_2_other)', 
                          "FisherPvalLess", "FisherPvalMore"]) + '\n'
        for mut_str_pair in self.mut_str_pair_to_obj:
            mut_pair = self.mut_str_pair_to_obj[mut_str_pair]
            if len(flatten(mut_pair.contingency_table)) < 4:
                continue
            if mut_pair.contingency_table[0][1] == 0:
                if mut_pair.contingency_table[1][0] == 0:
                    if mut_pair.contingency_table[0][0] < 2:
                        continue
            to_print = [mut_pair.name]
            to_print.append(str(len(mut_pair.phylo_edge_pairs)))
            to_print.append(str(mut_pair.contingency_table))
            to_print.append(str(mut_pair.p_val_less))
            to_print.append(str(mut_pair.p_val_more))
            out_str +=  '\t'.join(to_print) + '\n'
        return out_str

    

class CombinedPhyloMutPairs(object):

    def __init__(self, phylo_mut_pair_set, pval_thresh = 0.05):
        self.pval_thresh = pval_thresh
        self.mut_pair_to_tree_num = defaultdict(int)
        self.add_mut_pairs(phylo_mut_pair_set)
    

    def add_mut_pairs(self, phylo_mut_pair_set):
        '''
        add mutation pairs from a set calculated for a single tree
        all mutations passing the Fisher's exact test with a significant 
        right-tailed p-value are added
        '''
        for mut_str_pair in phylo_mut_pair_set.mut_str_pair_to_obj:
            if phylo_mut_pair_set.mut_str_pair_to_obj[mut_str_pair].p_val_more <= \
               self.pval_thresh:
                self.mut_pair_to_tree_num[mut_str_pair] += 1


    def write_phylo_network(self):
        out_str = ''
        for mut_pair in self.mut_pair_to_tree_num:
            out_str += '%s\t%s\t%d\n'%(mut_pair[0], mut_pair[1], 
                                    self.mut_pair_to_tree_num[mut_pair])
        

    def __repr__(self):
        out_str = ''
        for mut_pair in self.mut_pair_to_tree_num:
            out_str += '%s_%s\t%d\n'%(mut_pair[0], mut_pair[1], 
                                    self.mut_pair_to_tree_num[mut_pair])
        return out_str
        
               



# VBG: For better computational efficiency BioSeq, BioSeqAln
# can use numpy chararrays instead of the current dictionaries, or 
# an inbuilt PyCogent(GPL) sequence/aln object with modifications

from collections import OrderedDict
import copy
import re

from napa.utils.general import *
from napa.seq.parse import *
from napa.seq.format import *

class BioSeq(object):

    def __init__(self, seq_id = 'seqid', 
                 seq_str = '', 
                 seq_type = '', 
                 seq_pos_list = [],
                 seq_annot = {'':['']}):

        self.seq_id = seq_id
        self.seq_str = seq_str
        self.length = len(self.seq_str)

        if len(seq_type): 
            self.seq_type = seq_type 
        else:
            self.determine_seq_type(seq_str)

        self.assign_standard_chars()
        self.get_seq_pos(seq_pos_list)
        self.seq_annot = seq_annot


    def __repr__(self):
        return '\t'.join([str(el) for el in \
                          [self.seq_id, self.seq_str, self.seq_type, 
                           str(self.seq_annot)]])


    def add_annot(self, annot_dict):
        self.seq_annot.update(annot_dict)


    def get_seq_pos(self, seq_pos_list):
        if not len(seq_pos_list):
            stderr_write(['BioSeq WARNING: Initialized alignment positions\n',
                          '\tto default range 1..', len(self.seq_str), 
                          'for sequence with id:', self.seq_id])
            self.seq_pos_list = range(1, len(self.seq_str)+1)
        else:
            assert (len(seq_pos_list) == self.length), \
                'Lengths of sequence string and sequence position list ' + \
                'do not match: ' + ' '.join([str(el) \
                                             for el in [self.seq_id, self.length, 
                                                        len(seq_pos_list)]])
            self.seq_pos_list = seq_pos_list


    def extract_pos(self, pos_subset):
        seq_str = ''
        for posi, pos in enumerate(self.seq_pos_list):
            if pos in pos_subset:
                seq_str += self.seq_str[posi]
        self.seq_str = seq_str
        self.seq_pos_list = copy.deepcopy(pos_subset)
        self.length = len(self.seq_str)


    def determine_seq_type(self, seq_str):
        seq_char_set =  set(list(seq_str))
        if seq_char_set <= set(list('ATGC')):
            self.seq_type = 'DNA'
        elif seq_char_set <= set(list('AUGC')):
            self.seq_type = 'RNA'
        elif seq_char_set <= set(list('ACDEFGHIKLMNPQRSTVWY')):
            self.seq_type = 'Protein'
        else:
            self.seq_type = 'unknown'
        assert (self.seq_type in ['DNA','RNA','Protein']), \
            'Could not determine sequence type.' + \
            'Please enter standard Protein/DNA/RNA seq.'


    def assign_standard_chars(self):
        seq_type_to_chars = {'DNA':'ATGC', 'RNA':'AUGC', 
                             'Protein':'ACDEFGHIKLMNPQRSTVWY'}
        self.standard_chars = seq_type_to_chars[self.seq_type]


    def get_substitutions(self, other):
        '''
        Finds substitutions only (not insertions or deletions)
        between two sequences in alignment, considering 
        standard and unambiguous sequence characters only.
        (Needs proper error handling).
        '''

        if self.length != other.length or \
           self.seq_pos_list != other.seq_pos_list:
            stderr_write(['Need to align sequences before comparing'])
            return

        if self.seq_type != other.seq_type:
            stderr_write(['Sequences have distinct sequence types!'])
            stderr_write(['Can\'t compare sequences with IDS:', 
                          self.seq_id, 'and', other.seq_id])
            return 

        mut_list = []
        for i in range(self.length):
            if self.seq_str[i] != other.seq_str[i]:
                if self.seq_str[i] in self.standard_chars and \
                   other.seq_str[i] in self.standard_chars:
                    pos = str(self.seq_pos_list[i])
                    mut_list.append(self.seq_str[i] + pos + \
                                    other.seq_str[i])

        return mut_list



class BioSeqAln(object):
    
    def __init__(self, **kwargs):
        self.seq_type = kwargs.get('seq_type', 'Protein')
        self.seq_annot = kwargs.get('seq_annot', {'':''}) 
        self.aln_pos = kwargs.get('aln_pos', [])

        if len(kwargs.get('seqid_to_seq', {})):
            self.seqid_to_seq = copy.deepcopy(kwargs.get('seqid_to_seq', {}))
        elif len(kwargs.get('aln_fasta_file', '')):
            self.seqs_from_fasta_file(kwargs.get('aln_fasta_file', ''))

        if len(kwargs.get('annot_key', '')) and \
           (len(kwargs.get('seqid_to_annot_file', '')) or \
            len(kwargs.get('seqid_to_annot', {}))):
            self.annot_seqs(kwargs.get('annot_key', ''),
                            kwargs.get('seqid_to_annot_file', ''),
                            kwargs.get('seqid_to_annot', {}))
        
        if len(kwargs.get('sel_annot_key', '')) and \
            (kwargs.get('sel_annot_list', ['']) != [''] or \
            len(kwargs.get('sel_annot_file', ''))):
            self.subset_annot_seq_dict(sel_annot_key = kwargs.get('sel_annot_key', ''),
                                       sel_annot_file = kwargs.get('sel_annot_file', ''),
                                       sel_annot_list = kwargs.get('sel_annot_list', ['']))

        if len(kwargs.get('pos_subset', [])):
            self.subset_aln_pos(kwargs.get('pos_subset', []))

        if not len(self.aln_pos):
            stderr_write(['WARNING (BioSeqAln):',
                         'Default alignment positions set.'])
            self.aln_pos = self.seqid_to_seq[self.seqid_to_seq.keys()[0]].seq_pos_list

        self.depth = len(self.seqid_to_seq)
        self.length = len(self.aln_pos)



    def seqs_from_fasta_file(self, aln_fasta_file):
        self.seqid_to_seq = OrderedDict()
        self.add_sequences_from_file(aln_fasta_file)


    def add_sequences_from_file(self, aln_fasta_file):
        fasta_recs = fasta_iter(aln_fasta_file)
    
        for r in fasta_recs:
            seq_id, seq_str = r[0], r[1]
            seq_id = re.sub('[!@#$.]|','',seq_id)
            self.seqid_to_seq[seq_id] = BioSeq(seq_id = seq_id,
                                               seq_str = seq_str,
                                               seq_type = self.seq_type,
                                               seq_pos_list = self.aln_pos,
                                               seq_annot = \
                                               copy.deepcopy(self.seq_annot))


    def annot_seqs(self, annot_key = '', seqid_to_annot_file = '',
                   seqid_to_annot = {}):
        '''
        Add sequence annotation to sequences in alignment.
        '''
        if len(seqid_to_annot_file):
            aln_seqid_to_annot = parse_keyval_dict(seqid_to_annot_file)
        elif len(seqid_to_annot):
            aln_seqid_to_annot = seqid_to_annot
        else:
            stderr_write('WARNING (BioSeqAln): No functions assigned')
        for seqid in aln_seqid_to_annot:
            seqid_fasta = re.sub('[!@#$.]|','', seqid)
            if seqid_fasta in self.seqid_to_seq:
                self.seqid_to_seq[seqid_fasta].add_annot(\
                                {annot_key: aln_seqid_to_annot[seqid]})


    def subset_aln_pos(self, pos_subset = []):
        if not len(pos_subset):
            return # No subsetting of columns
        for seqid in self.seqid_to_seq:
            self.seqid_to_seq[seqid].extract_pos(pos_subset)
        self.aln_pos = pos_subset
        self.length = len(pos_subset)


    def seqids_with_annot(self, annot_key, annot_list):
        seqid_annot = []
        for seqid in self.seqid_to_seq:
            if annot_key not in self.seqid_to_seq[seqid].seq_annot:
                continue
            if self.seqid_to_seq[seqid].seq_annot[annot_key] in annot_list:
                seqid_annot.append(seqid)
        return seqid_annot

                                                         
    def subset_annot_seq_dict(self,sel_annot_key = '', sel_annot_file = '', 
                              sel_annot_list = ['']):
                                        
        '''
        Subset alignment sequences by a sequence id 
        using a sequence dictionary based on sequence annotation
        e.g. only sequences having a given function.
        '''
        if sel_annot_list != [''] and len(sel_annot_file):
            sel_annot_list = parse_column(sel_annot_file)

        seqid_to_seq_annot = OrderedDict()
        for seqid in self.seqids_with_annot(sel_annot_key, sel_annot_list):
            seqid_to_seq_annot[seqid] = copy.deepcopy(self.seqid_to_seq[seqid])
        
        self.seqid_to_seq.clear()
        self.seqid_to_seq = copy.deepcopy(seqid_to_seq_annot)


    def get_seq_muts(self, wt_seq = ''):
        '''
        Obtain mutations between each alignment sequence and
        a reference sequence.
        '''
        self.seqid_to_mut = OrderedDict()
        for seqid in self.seqid_to_seq:
            self.seqid_to_mut[seqid] = wt_seq.get_substitutions(\
                                                    self.seqid_to_seq[seqid])

    def get_seq_muts_id(self, wt_id = ''):
        
        try:
            assert wt_id in self.seqid_to_seq
            wt_seq = self.seqid_to_seq[wt_id]
            self.get_seq_muts(wt_seq)
        
        except AssertionError:
            stderr_write(['BioSeqAln ERROR: Sequence id not in',
                          'alignment. Cannot extract mutations!'])


    def get_seq_muts_id_seq(self, wt_id = '', wt_seq_str = ''):
        
        if len(wt_id) and len(wt_seq_str):
            if wt_id in self.seqid_to_seq:
                if self.seqid_to_seq[wt_id].seq_str == wt_seq_str:
                    self.get_muts_id(wt_id)
                else:
                    stderr_write(['BioSeqAln WARNING:',
                                  'WT present in alignment but',
                                  'with different id.'])
                    wt_seq = BioSeq(wt_id, wt_seq_str, self.aln_pos)
                    self.get_seq_muts(wt_seq)
            else:
                wt_seq = BioSeq(wt_id, wt_seq_str, self.aln_pos)
                self.get_seq_muts(wt_seq)

        elif not len(wt_id) and len(wt_seq_str):
            wt_id = 'WildType'
            wt_seq = BioSeq(wt_id, wt_seq_str, self.aln_pos)
            self.get_seq_muts(wt_seq)

        elif len(wt_id) and not len(wt_seq_str):
            self.get_seq_muts_id(wt_id)

    def __repr__(self):
        out_str = ''
        for seqid in self.seqid_to_seq:
            out_str += write_wrap_fasta_seq(seqid+ '___' + '__'.join(\
                                                        ['_'.join(ann_pair) for ann_pair in \
                                                         self.seqid_to_seq[seqid].seq_annot.items() \
                                                         if '' not in ann_pair]), 
                                            self.seqid_to_seq[seqid].seq_str)
        return out_str

    def write_muts(self, out_file, pos_subset):
        with open(out_file, 'wb') as of:
            for seqid in self.seqid_to_mut:
                if len(pos_subset):
                    of.write('\t'.join(seqid, 
                                   ','.join([mut for mut in self.seqid_to_mut[seqid] \
                                             if mut in pos_subset]))+'\n')
                else:
                    of.write('\t'.join(seqid, ','.join(self.seqid_to_mut[seqid]))+'\n')






import argparse
import sys
import itertools
import sys

#print sys.path

from napa.utils.general import *
from napa.mutpair.aln import * # Netw. construct from alignment


def parse_args():
        parser = argparse.ArgumentParser()

        parser.add_argument('-a', dest = 'alnFASTA', required = True, 
                            help = 'Path of multiple sequence alignment fasta file.')

        parser.add_argument('-p', dest = 'posList', required = True, 
                            help = 'File containing protein residue positions ' + \
                            'corresponding to each column in the alignment.')

        parser.add_argument('-wt', dest = 'wtId', required = True,  
                            help = 'Id of wildtype or ancestral sequence, ' + \
                            'assuming it is already in alignment. Can also provide '+ \
                            'separate FASTA with wildtype sequence.')

        parser.add_argument('-ps', dest = 'posSubset', required = True, 
                            help = 'File containing protein positions' + \
                            'to be considered in network')

        parser.add_argument('-pf', dest = 'protFunc', required = True, 
                            help = 'File containing known protein functional annot. ' + \
                            'of sequences in alignment.')

        parser.add_argument('-sf', dest = 'selProtFunc', required = True, 
                            help = 'String/File containing protein funcs' + \
                            'to be included in net.')

        parser.add_argument('-nf', dest = 'netFile', required = True, 
                            help = 'Outputfile for result network.')

        args = parser.parse_args()

        return args

class AlnNetInput(object):
        '''
        Prepare input files for alignment-based network reconstruction.
        '''
        def __init__(self, args):

                # List of positions/column numbers in alignment
                # If not provided, positions start with 1,2..alignment_length
                self.pos_list = [int(rec) for rec in parse_column(args.posList)]


                # List of protein residue positions considered for network.
                # (remove residues from signaling peptide, or other domains not studied).
                self.pos_subset = [int(rec) for rec in parse_column(args.posSubset)]


                # prot. func. for each seq. in aln with known func. annot.
                self.seqid_to_prot_func = parse_keyval_dict(args.protFunc)

                # subset of protein selected functions considered 
                self.sel_prot_func_list = parse_column(args.selProtFunc)


def main():
    args = parse_args()

    # Construct network based on alignment
    inp = AlnNetInput(args)
    alnMutPairSet = AlnMutPairSet(aln_fasta_file = args.alnFASTA,
                                  aln_pos = inp.pos_list,
                                  wt_id = args.wtId,
                                  pos_subset = inp.pos_subset, 
                                  seqid_to_prot_func = inp.seqid_to_prot_func,
                                  sel_prot_func = inp.sel_prot_func_list)
        
    with open(args.netFile, 'wb') as f:
        f.write(alnMutPairSet.write_jaccard_weights_network(min_co_occur = 2))
    
if __name__ == '__main__': main()





import argparse
import sys
import os.path
from collections import defaultdict
from itertools import islice
from copy import deepcopy
import math
import random

import networkx as nx
from networkx.utils import weighted_choice



from napa.utils.general import *


def window(seq, n=2):
    '''
    Returns a sliding window (of width n) over data 
    from the iterable:
    s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ...                   "
    '''
    it = iter(seq)
    result = tuple(islice(it, n))
    if len(result) == n:
        yield result    
    for elem in it:
        result = result[1:] + (elem,)
        yield result

def file_line_iterator(file_name):
    with open(file_name, 'rb') as f:
        for line in f: 
            yield line.strip()

def file_line_list(file_name):
    with open(file_name, 'rb') as f:
        return f.readlines()

#-----------------------------------------------------------------------#
class MutNet(object):

    '''
    Mutation network analysis methods
    '''

    def __init__(self, net = None, net_file = None, 
                 net_type = None):
        '''
        Networks based on tab-delimited files
        source\ttarget\tweight
        weight: weights from alignment/phylo networks
        net_type: 'directed' or 'undir'ected
        '''
        if net != None:
            self.from_nx_net(net, net_type)
        else:
            self.read_net(net_file, net_type)

        # inverted edges net for some centralities
        self.get_reverse_net()
        # inverted weights for centralities requiring distance
        self.get_inv_w()
        

    def from_nx_net(self, net, net_type):
        try:
            self.g = deepcopy(net)
            self.directed = nx.is_directed(net)
        except:
            raise Exception('NetworkX format error.')


    def get_reverse_net(self):
        if self.directed:
            self.rg = nx.reverse(self.g) 
        else:    
            self.rg = deepcopy(self.g)
   

    def read_net(self, net_file, net_type):
        '''
        Read from file
        Create networkX graph object (directed or undirected)
        '''
        if net_file == None or not os.path.isfile(net_file):
            stderr_write(['WARN: Missing net file. '+ \
                          'Empty network created.'])
            if net_type == 'undir':
                self.g = nx.Graph()
                self.directed = False
            else:
                self.g = nx.DiGraph()
                self.directed = True
            return

        if net_type == 'undir':
            self.g = nx.parse_edgelist(file_line_iterator(net_file), 
                                       data=(('weight',float),), 
                                       create_using=nx.Graph())
            self.directed = False
        
        else:
            self.g = nx.parse_edgelist(file_line_iterator(net_file), 
                                       data=(('weight',float),), 
                                       create_using=nx.DiGraph())
            self.directed = True

#-----------------------------------------------------------------------#
    def get_inv_w(self):
        '''
        Weight inversion to convert weights of association to
        distances between linked nodes. 
        The higher the association, the lower the distance.
        Used in all shortest paths and path length methods.
        '''

        if not len(self.g): return
        if not len(nx.get_edge_attributes(self.g, 'weight').values()):
            return
        max_weight = float(max([d['weight'] \
                                for u, v, d in self.g.edges(data=True)]))
        min_weight = float(min([d['weight'] \
                                for u, v, d in self.g.edges(data=True)]))


        nx.set_edge_attributes(self.g, 'weight',
                              {(u, v):d['weight']/max_weight \
                               for u, v, d in self.g.edges(data=True)})

        nx.set_edge_attributes(self.g, 'invWeight',
                              {(u, v):(min_weight/max_weight)/d['weight'] \
                               for u, v, d in self.g.edges(data=True)})


        nx.set_edge_attributes(self.rg, 'weight',
                              {(u, v):d['weight']/max_weight \
                               for u, v, d in self.rg.edges(data=True)})

        nx.set_edge_attributes(self.rg, 'invWeight',
                              {(u, v):(min_weight/max_weight)/d['weight'] \
                               for u, v, d in self.rg.edges(data=True)})
            
    def get_default_cent_list(self):
        '''
        Get all centralities that can be calculated based on 
        applicability to directed or undirected graphs.
        '''
        if nx.is_directed(self.g):
            self.node_cent_list =  ['loc.in.deg', 'loc.out.deg', 
                                    'loc.in.strength', 
                                    'loc.out.strength',
                                    'glob.close', 'glob.eigen',
                                    'glob.pagerank', 'glob.betw', 
                                    'glob.kpath']
        else:
            # glob.info removed from list
            self.node_cent_list = ['loc.deg', 'loc.strength', 'glob.close', 
                                   'glob.eigen', 
                                   'glob.pagerank', 'glob.betw', 'glob.kpath']

        return self.node_cent_list



    def normalize(self, node_att_dict = {}, norm_factor = 1.):
        norm_factor = 1. if norm_factor == 0. else norm_factor
        d = defaultdict(float)
        for k in node_att_dict:
            d[k] = float(node_att_dict[k])/norm_factor
        return d

#-----------------------------------------------------------------------#
    def get_node_centralities(self, cent_list = None):
        '''
        Combined standard (local and shortest-path)
        and kpath centrality metrics for 
        single nodes. 
        '''
        if cent_list == None:
            self.get_default_cent_list()
        else:
            self.node_cent_list = cent_list

        if 'loc.in.deg' in self.node_cent_list:
            norm_factor = self.g.number_of_nodes() - 1.
            # number of in neighbors (disreguard link weights)
            # normalize each node by remaining number of nodes in network
            nx.set_node_attributes(self.g, 'loc.in.deg', 
                                   self.normalize(self.g.in_degree(weight=None),
                                             norm_factor))

        if 'loc.out.deg' in self.node_cent_list:
            norm_factor = self.g.number_of_nodes() - 1.
            nx.set_node_attributes(self.g, 'loc.out.deg', 
                                   self.normalize(self.g.out_degree(weight=None),
                                                  norm_factor))
        

        if 'loc.deg' in self.node_cent_list:
            norm_factor = self.g.number_of_nodes() - 1.
            nx.set_node_attributes(self.g, 'loc.deg', 
                                   self.normalize(self.g.degree(weight=None),
                                                  norm_factor))

        if 'loc.strength' in self.node_cent_list:
            norm_factor = self.g.number_of_nodes() - 1.
            nx.set_node_attributes(self.g, 'loc.strength', 
                                   self.normalize(self.g.degree(weight = 'weight'),
                                                  norm_factor))

        if 'loc.in.strength' in self.node_cent_list:
            norm_factor = self.g.number_of_nodes() - 1.            
            nx.set_node_attributes(self.g, 'loc.in.strength', 
                                   self.normalize(self.g.in_degree(weight = 'weight'),
                                                  norm_factor))

        if 'loc.out.strength' in self.node_cent_list:
            norm_factor = self.g.number_of_nodes() - 1. 
            nx.set_node_attributes(self.g, 'loc.out.strength',
                                   self.normalize(self.g.out_degree(weight = 'weight'),
                                                  norm_factor))

        # Note: all (shortest) path-length based centralities use the 
        #inverse weight -- this is a proxy for distance, 
        #s.t. stronger association weight = shorter dist.
        if 'glob.close' in self.node_cent_list:
            nx.set_node_attributes(self.g, 'glob.close', 
                                   nx.closeness_centrality(self.g, 
                                                           distance = 'invWeight', 
                                                           normalized = True))
        if 'glob.info' in self.node_cent_list:
            nx.set_node_attributes(self.g, 'glob.info',
                                   nx.current_flow_closeness_centrality(self.g, 
                                                                        weight = 'weight'))
        if 'glob.eigen' in self.node_cent_list:
            nx.set_node_attributes(self.g, 'glob.eigen',
                                   nx.eigenvector_centrality(self.rg, weight = 'weight'))

        if 'glob.pagerank' in self.node_cent_list:
            nx.set_node_attributes(self.g, 'glob.pagerank',
                                   nx.pagerank(self.rg, weight = 'weight'))

        # betweenness is not distance based in networkx
        if 'glob.betw' in self.node_cent_list:
            nx.set_node_attributes(self.g, 'glob.betw',
                                   nx.betweenness_centrality(self.g, weight = 'weight', 
                                                             normalized = True))

        if 'glob.kpath' in self.node_cent_list:
            node_kpath_cents = self.path_kpath_centrality(alpha = 0.)

            for node in self.g.nodes():
                if str(node) in node_kpath_cents:
                     self.g.node[node]['glob.kpath'] = node_kpath_cents[str(node)]
                else: 
                    self.g.node[node]['glob.kpath'] = 0.


  #-----------------------------------------------------------------------#      
    def str_node_centralities(self, header = False, 
                              prefix = '', suffix = '',
                              nodes = None, sorted_cent_names = None):
        '''
        Write centralities for (individual) nodes in the network.
        '''
        
        if sorted_cent_names == None:
            sorted_cent_names = \
                flatten([self.g.node[node].keys() \
                         for node in self.g.nodes_iter()])
            sorted_cent_names = sorted(list(set(sorted_cent_names)))
        
        out_list = []
        if header:
            header = 'node.or.path\t' + (prefix + suffix) + '\t' + \
                     '\t'.join(sorted_cent_names)+'\tnum.net.nodes'
            out_list += [header]

        if nodes != None:
            for node in set(nodes) & set(self.g.nodes()):
                out_list +=  ['%s\t%s\t%s\t'%(prefix, str(node), suffix) + \
                              '\t'.join(\
                              ['{:.4g}'.format(self.g.node[node][cent]) \
                               for cent in sorted_cent_names]) + '\t' + \
                              str(self.g.number_of_nodes())]
        else:
            for node in self.g.nodes():
                out_list += [prefix + str(node) + suffix + \
                             '\t'+ '\t'.join(\
                             ['{:.4g}'.format(self.g.node[node][cent]) \
                              for cent in sorted_cent_names]) + '\t' + \
                             str(self.g.number_of_nodes())]

        if not len(out_list): return ''

        return '\n'.join(out_list) + '\n'


  #-----------------------------------------------------------------------#      
    def get_desc_net_node_cent(self, source = None, exclude_nodes = [],
                               cent_list = None):
        nodes_in_graph =  nx.descendants(self.g, source) | set([source])
        nodes_in_graph -=  set(exclude_nodes)
        if not len(nodes_in_graph): 
            desc_net = MutNet()
        desc_net = MutNet(net = self.g.subgraph(nodes_in_graph))
        desc_net.get_node_centralities(cent_list = cent_list)
        return desc_net


    def get_rel_cent(self, cent_list = None, prev_nodes = [], 
                     nodes = None, path_len = 3):
        
        if len(prev_nodes) >= path_len: return ''
        
        if nodes != None and not len(nodes): nodes = None

        
        prefix = '\t'.join(prev_nodes)
        suffix = (path_len - len(prev_nodes) - 1)*'\t'
        
        self.get_node_centralities(cent_list = cent_list)
        out_str = self.str_node_centralities(prefix = prefix,
                                             suffix = suffix,
                                             nodes = nodes)

        for node in self.g:

            if node in prev_nodes: continue
            desc_net = self.get_desc_net_node_cent(source = node,
                                                   exclude_nodes = prev_nodes,
                                                   cent_list = cent_list)
            if desc_net.g.size() < 3: continue

            out_str += desc_net.get_rel_cent(\
                        cent_list = cent_list, prev_nodes = prev_nodes + [node],
                        nodes = list(set(self.g.neighbors(node)) - set(prev_nodes)))

        return out_str


    def get_path_between_path_cent(self, path_node_length = 2,
                                          cent_list = ['glob.betw', 'glob.kpath']):
        '''
        Get shortest and k-path betweenness centralities for all 
        paths consisting of path_node_length number of nodes.
        Here path_len refers to number of nodes rather than edges in path.
        '''
        self.path_cent = {}
        
        if 'glob.betw' in cent_list:
            self.path_shortest_path_b_cent(path_len = path_node_length)

        if 'glob.kpath' in cent_list:            
            self.path_kpath_centrality(alpha = 0., path_len = path_node_length)

            
    def write_path_betw_path_cent(self, output_file):
        with open(output_file, 'wb') as f:
            sorted_cent_names = sorted(self.path_cent.keys())
            f.write('\t'.join(['Path']+[cent for cent in sorted_cent_names]) + '\n')
            
            path_set = set()

            for path_dict in self.path_cent.values():
                path_set |= set(path_dict.keys())

            for path in path_set:
                f.write(path + '\t' + \
                        '\t'.join([str(self.path_cent[cent][path]) \
                                   if path in self.path_cent[cent] \
                                   else str(0.) for cent in sorted_cent_names]) + '\n')

    #def get_gn_communities(self): 
        #self.components =  girvan_newman(self.g, weight='weight')
        


    def write_communities(self, output_file):
        '''
        Not tested: check that mutations are output rather than
        node numbers from networkX.
        '''
        with open(output_file, 'wb') as f:
            f.write('Node\tcommunity_label\n')

            for comp_i, component in enumerate(self.components):
                for node_name in component:
                    f.write('%s\t%d\n'%(node_name, comp_i + 1)) 
                                            

    def all_sps(self, source, target):
        '''
        No error all_shortest_paths between source and target
        '''
        try:
            return [p for p in nx.all_shortest_paths(self.g, source, target,
                                                     weight = 'invWeight')]
        except nx.NetworkXNoPath:
            return []


    def path_shortest_path_b_cent(self, path_len = 2):
        '''
        Extension of the single-node shortest path (sp) betweenness
        to a set of contiguous nodes in the network
        can be further optimized
        '''
        num_all_paths = 0
        self.path_cent['shortest_path'] = defaultdict(float)

        for source in self.g.nodes():
            for target in self.g.nodes():
                if source == target: 
                    continue #suboptimal, better iteration strategy?
                st_paths = self.all_sps(source, target)    
                num_all_paths += len(st_paths)
                for st_path in st_paths:
                    sub_paths = [wn for wn in window(st_path[1:-1], n = path_len)]
                    for sub_path in sub_paths:
                        sub_path_str_list = [str(node) for node in sub_path]

                        if path_len == 1:
                            self.path_cent['shortest_path']['_'.join(sub_path_str_list)] += 2
                            continue

                        if not self.directed:
                            self.path_cent['shortest_path']['_'.join(\
                                                        reversed(sub_path_str_list))] += 1
                            self.path_cent['shortest_path']['_'.join(sub_path_str_list)] += 1

                        else:
                            self.path_cent['shortest_path']['_'.join(sub_path_str_list)] += 2

        for path in self.path_cent['shortest_path']:
            self.path_cent['shortest_path'][path] *= (0.5 / num_all_paths)



    def path_kpath_centrality(self, k = None, alpha = 0.2, 
                              weight = 'invWeight', seed = 123456, path_len = 1):
        '''
        Adapted from Alakahoon et al. and published code
        by extending from single node to path centrality
        '''

        kpath_path_cent = defaultdict(float)

        if self.g.is_multigraph():
            raise nx.NetworkXError("Not implemented for multigraphs.")

        n = self.g.number_of_nodes()
        ne = self.g.number_of_edges()

        if n <= path_len + 4 or ne < path_len + 4:
            # network is too small for calculating path centr. of this length
            return {}
            
        if k is None: 
            k = int(math.log(n + ne))

        if k < path_len + 4: k = min(path_len + 4, n)
            
        T = 2. * k**2 * n**(1-2*alpha) * math.log(n)
        
        random.seed(seed)

        for i in range(int(T+1)):
            st_path = []
            s = random.choice(self.g.nodes()) # choose source node
            l = random.randint(path_len + 2, k) # choose a random path length
            st_path.append(s)

            for j in range(l): # fill out a path of length l
                #invert distance again - get association
                nbrs = {nbr: 1./d.get(weight, 1.0) \
                        for nbr,d in self.g[s].items() \
                        if nbr not in st_path} 
                if not nbrs: break
                v = weighted_choice(nbrs)
                st_path.append(v)
                s = v # set the current path source (current node) to v

            if len(st_path) < path_len + 4: continue

            # sub_paths exclude end-points
            sub_paths = [wn for wn in window(st_path[1:-1], n = path_len)]
            for sub_path in sub_paths:
                sub_path_str_list = [str(node) for node in sub_path] 

                if path_len == 1:
                    kpath_path_cent['_'.join(sub_path_str_list)] += 2.
                    continue

                if not self.directed:
                    kpath_path_cent['_'.join(reversed(sub_path_str_list))] += 1.
                    kpath_path_cent['_'.join(sub_path_str_list)] += 1.
                else:
                    kpath_path_cent['_'.join(sub_path_str_list)] += 2.


        for path in kpath_path_cent:
            kpath_path_cent[path] *= (0.5 / T)

        return kpath_path_cent



#! /usr/bin/env python 
'''Custom functions for writing sequence files. 
Should be replaced with optimized inbuilt functions from PyCogent(GPL).'''

def write_wrap_fasta_seq(seqId,  seqStr,  wrap_at = 80,  
                         endline = '\n'):
    ''' 
    Print FASTA sequence and wrap at desired character count 
    '''
    s = []
    s.append('>%s%s' % (seqId.lstrip('>'),  endline))
    exploded_seq = list(seqStr)
    wrap_points = range(0, len(exploded_seq), wrap_at)
    wrap_points.reverse()
    for i in wrap_points[:-1]:
        exploded_seq.insert(i,  endline)
    s = s + exploded_seq + [endline]
    return ''.join(s)


def write_wrap_fasta_dict(id_sequence_dict={}, 
                          fasta_output_name='fasta_dict.fa',
                          wrap_at=80, sort=False):
    ''' 
    Prints FASTA sequences from id vs. sequence dictionary 
    allows for sorting fasta records alphabetically by fasta id.
    '''
    with open(fasta_output_name,  'w') as f:
        if not sort:
            for seq_id in id_sequence_dict:
                f.write(write_wrap_fasta_seq(seq_id, 
                                         id_sequence_dict[seq_id], 
                                         wrap_at))
        else:
            for seq_id in sorted(id_sequence_dict.keys()):
                f.write(write_wrap_fasta_seq(seq_id, 
                                         id_sequence_dict[seq_id], 
                                         wrap_at))
                                        
            



#! /usr/bin/env python 

from ete2 import PhyloTree
from napa.seq.BioSeq import *


class PhyloEdge(object):
    ''' 
    PhyloEdge line graph of PhyloTree from ete2.
    PhyloEdge nodes are egdes in the PhyloTree
    Edges are oriented such that parental node always
    comes before child node.
    Sequences on nodes are BioSeq objects.
    '''

    def __init__(self, parent_node = None, child_node = None,
                 aln = None, wt_seq = None,
                 parent = None, children = []):
        
        # parent node in edge, a PhyloTree Object        
        self.parent_node = parent_node 
        self.parent_node_name = "ANC" if not parent_node else parent_node.name

        # child node in PhyloTree, representing edge target
        self.child_node = child_node
        self.child_node_name = child_node.name

        # name edge by parent-child names
        self.name = '%s__%s' % (self.parent_node_name, self.child_node_name)

        # assign BioSeq sequence objects to parent and child 
        self.wt_seq = wt_seq
        self.aln = aln
        self.assign_node_seqs()

        # Get mutations for all sequences in alignment
        self.aln.get_seq_muts(wt_seq = self.wt_seq)        
        
        # Get set of mutations/mut positions along this PhyloTree edge
        self.get_muts()
        self.mut_pos = [get_int(m) for m in self.muts]

        # parent edge = another node pair whose target = phyloEdge's source
        self.parent = parent 

        # child edges = node pairs whose source = phyloEdge's target
        self.get_node_children(children)
                
        self.precs = [] 
        self.followers = []
        self.ordered_precs = []


    def __iter__(self):
        for p_edge in chain(*imap(iter, self.children)):
            yield p_edge
        yield self


    def assign_node_seqs(self):
        if self.parent_node_name in self.aln.seqid_to_seq:
            self.parent_node_seq = self.aln.seqid_to_seq[self.parent_node_name]
        else:
            self.parent_node_seq = self.wt_seq

        if self.child_node_name in self.aln.seqid_to_seq:
            self.child_node_seq = self.aln.seqid_to_seq[self.child_node_name] 
        else:
            self.child_node_seq = self.wt_seq
                

    def add_child(self, phylo_edge):
        if phylo_edge.name not in [c.name for c in self.children]:
            self.children.append(phylo_edge)
            phylo_edge.parent = self


    def get_node_children(self, children):
        if len(children):
            self.children = children
            return
        self.children = []
        child_node_children = self.child_node.get_children()
        for child_node_child in child_node_children:
            self.add_child(\
                    PhyloEdge(parent_node = self.child_node,
                              child_node = child_node_child,
                              aln = self.aln, wt_seq = self.wt_seq,
                              parent = self))


    def build_edge_tree(self):
        self.children = [child for child in self.children \
                         if child.parent.name == self.name]
        self.build_edge_followers() 
        self.build_edge_precs()


    def build_edge_followers(self):
        for child in self.children:
            child.build_edge_followers()
            self.followers  +=  [child] + child.followers
        self.followers = list(set(self.followers))


    def build_edge_precs(self):
        for child in self.children:
            child.precs = list(set(self.precs + [self]))
            child.build_edge_precs()

   
    def assign_ordered_precs(self):
        '''
        PhyloEdge's precursor edges in order on tree.
        '''
        if not len(self.precs): 
            self.ordered_precs = []
            return

        parent_to_child_node = {}
        for p in self.precs:
            p_name = p.parent_node.name if p.parent_node else 'ANC'
            np_name = p.child_node.name if p.child_node else 'ANC'
            parent_to_child_node[p_name] = np_name

        most_anc_node_name = list(set(parent_to_child_node.keys()) - \
                           set(parent_to_child_node.values()))

        if not(len(most_anc_node_name)): 
            stderr_write(['No ancestor found for phylogeny edge:', self.name])

        if len(most_anc_node_name) > 1: 
            stderr_write(['Found multiple ancestors for phylo edge:', self.name])
        
        curr_anc_name = most_anc_node_name[0]
        max_depth = len(set(parent_to_child_node.keys() + \
                            parent_to_child_node.values()))

        for i in range(max_depth - 1):
            next_anc_name = parent_to_child_node[curr_anc_name]
            next_prec = [p for p in self.precs if p.name == \
                             curr_anc_name+'__'+next_anc_name]
            self.ordered_precs.append(next_prec[0])
            curr_anc_name = next_anc_name

        if self.ordered_precs[-1] != self.parent or \
           len(self.ordered_precs)!= len(self.precs):
            print "_precs not properly ordered/found:", self.name
            print "\tparent_to_child_node", parent_to_child_node
            print "\tordered", [p.name for p in self.ordered_precs if p!=None]
            print "\tunordered", [p.name for p in self.precs]


    def get_common_precs(self, other, annot_key, annot_combinations):
        '''
        Common precursors for a two phylo edges. 
        '''
        common_prec = []
        if self.parent == None or other.parent == None:
            common_prec = [] # case of one of the edges being root
        elif self ==  other:
            common_prec = [self, self.parent]
        elif self in other.precs:
            common_prec = [self.parent, self]
        elif other in self.precs:
            common_prec = [other.parent, other]
        else: # they are in different lineages: look for Mrca
            common_prec = list(set(self.precs) & set(other.precs))
            if not len(common_prec):
                print "No common precs for", self.name, other.name
            elif len(common_prec) > 1:
                common_prec = [self.ordered_precs[len(common_prec)-1]]
        common_prec = list(set([p for p in common_prec \
                                if p.check_node_seq_annot(annot_key, annot_combinations)]))
        return common_prec


    def get_intermediate_prec(self, prec_list, annot_key, annot_combinations):
        '''
        Intermediate edges between two phylo edges.
        '''
        intermediates = [self] if self.check_node_seq_annot(annot_key, annot_combinations) \
                        else []

        if not len(prec_list):
            intermediates  +=  [p for p in self.ordered_precs \
                              if p.check_node_seq_annot(annot_key, annot_combinations)]
            return intermediates

        first_prec = self.ordered_precs.index(\
                                        next((p for p in self.ordered_precs \
                                              if p in prec_list), 
                                             self.ordered_precs[-1]))

        return intermediates + [im for im in self.ordered_precs[:first_prec] \
                                if im.check_node_seq_annot(annot_key, annot_combinations)]
        

    def get_muts(self):
        ''' 
        Get mutations connecting two phylogeny nodes (sequences)
        Mutations from a reference sequence are considered
        Overlap is removed.
        '''
        self.muts = []
        # Get mutations from a reference sequence for parent/child node
        # Those should have been pre-calculated in self.aln from external function
        # creating the PhyloEdge instance
        
        parent_node_mutations, child_node_mutations = [], []
        if self.parent_node_name in self.aln.seqid_to_mut:
            parent_node_mutations = self.aln.seqid_to_mut[self.parent_node_name]
        if self.child_node_name in self.aln.seqid_to_mut:
            child_node_mutations = self.aln.seqid_to_mut[self.child_node_name]
        self.muts = list(set(child_node_mutations) - set(parent_node_mutations))    


    def is_leaf_edge(self):
        return self.child_node.is_leaf()


    def check_dist(self, other, dist_thresh):
        return self.parent_node.get_distance(other.child_node) <=  dist_thresh


    def check_node_seq_annot(self, annot_key, annot_combinations):
        if annot_key not in self.parent_node_seq.seq_annot:
            return False
        if annot_key not in self.child_node_seq.seq_annot:
            return False
        
        parent_node_annot = self.parent_node_seq.seq_annot[annot_key]
        child_node_annot = self.child_node_seq.seq_annot[annot_key]

        if parent_node_annot in annot_combinations:
            if child_node_annot in annot_combinations[parent_node_annot]:
                return True
        return False


    def check_mut_pos(self):
        if not len(self.mut_pos): return False
        return set(self.mut_pos) <= self.aln.aln_pos


    def get_newick(self):
        ''' 
        Get newick format string for phylo edge tree.
        '''

        if self.parent ==  root_parent:
            if not len(self.children): 
                return nwk_str + str(self.name)
            child_names = [c.name for c in self.children]
            child_names.sort()
            children_str = ', '.join([str(c) for c in child_names])
            if str(self.name) not in re.findall(r"[\w'] + ", nwk_str):
                nwk_str += '(' + children_str + ')' + str(self.name)
                for child in self.children:
                    nwk_str = child.get_newick(nwk_str)
                else:
                    for child in self.children:
                        nwk_str = child.get_newick(nwk_str)
        
        if not len(self.children): return nwk_str

        possible_strings = ['(' + str(self.name) + ')']
        possible_strings  +=  ['(' + str(self.name) + ', ']
        possible_strings  +=  [', ' + str(self.name) + ', ']
        possible_strings  +=  [', ' + str(self.name) + ')']

        child_names = sorted([c.name for c in self.children])
        children_str = ', '.join([str(c) for c in child_names])

        for p_str in possible_strings:
            if p_str in nwk_str:
                nwk_str = nwk_str.replace(p_str, 
                                          p_str[0] + '(' + children_str + ')' +  \
                                          str(self.name) + p_str[_1])
                break

            ordered_children = sorted(self.children, key = lambda c: c.name)

            for child in ordered_children:
                nwk_str = child.get_newick(nwk_str)

        return nwk_str


    def __repr__(self):
        to_print = [self.name, 'mutations: ' + ';'.join(self.muts), 
                   'children: ' + ';'.join([c.name for c in self.children]), 
                   'followers: ' + ';'.join([f.name for f in self.followers]), 
                   'precs: ' + ';'.join([p.name for p in self.precs])]
        if len(self.ordered_precs):
            to_print.append('ordered_prec: ' + \
                            ';'.join([op.name \
                                      for op in self.ordered_precs if op])) 
        return '\n\t'.join(to_print)
                                         



        
        
    



import argparse
import os, sys

pwd = os.getcwd()
sys.path.append(pwd[0:pwd.find('/NAPA/')]) # NAPA package path


from napa.utils.general import *
from napa.net.net import *

def parseArgs():
    #Notes: perhaps better a yaml file with options for analysis
    parser = argparse.ArgumentParser( \
    description = 'Characterize mutation network and its nodes/links/clusters.')
    
    parser.add_argument('-nf', dest = "netFile", type=str, required=True,
                        help = 'Path to where weighted mutation'+\
                        ' pairs are stored.')

    parser.add_argument('-mt', dest = "mutNetType", type=str, required=True,
                        help = 'Type of mutation pair network: directed or undirected.')
    
    parser.add_argument('-pl', dest = "pathLength", type=int, default=2,
                        help = 'Length of central paths to consider, in number'+\
                        ' of nodes.')

    parser.add_argument('-co', dest = "centOutFile", type=str, 
                        default='net_centralities.txt',
                        help = 'Output file for centralities.')

    parser.add_argument('-ct', dest = "centralityType", type=str, 
                        default='net_path_centralities.txt',
                        help = 'node, path, or relative centrality.')


    args = parser.parse_args()
    return args

                
def main():
    args = parseArgs()
    mn = MutNet(net_file=args.netFile, net_type = args.mutNetType)
    
    cent_type = args.centralityType
    if cent_type == 'node':
        mn.get_node_centralities()
        with open(args.centOutFile, 'wb') as cf:
            cf.write(mn.str_node_centralities(header = True))
            
    elif cent_type == 'path':
        mn.get_path_between_path_cent(path_node_length = args.pathLength)
        mn.write_path_betw_path_cent(args.centOutFile)

    elif 'rel' in cent_type: 
        if cent_type == 'rel':
            # all relative centralities
            centralities =  mn.get_default_cent_list()

        elif cent_type == 'relloc':
            # local centrality types only: degree, clustering...
            centralities =  [c for c in mn.get_default_cent_list() if 'loc' in c]

        elif cent_type == 'relglob':
            # global centrality types only: closeness, eigenvalue, betweenness...
            centralities =  [c for c in mn.get_default_cent_list() if 'glob' in c]


        header = 'node.or.path' + (args.pathLength - 1) * '\t' + '\t'.join(centralities) + \
                 '\tnum.net.nodes\n'
        out_str = mn.get_rel_cent(cent_list = centralities, path_len = args.pathLength)

        with open(args.centOutFile, 'wb') as cf:
            cf.write(header + out_str)

if __name__ == "__main__": main()



# -*- coding: utf-8 -*-
'''
Copyright (C) 2004-2012, NetworkX Developers
Aric Hagberg <hagberg@lanl.gov>
Dan Schult <dschult@colgate.edu>
Pieter Swart <swart@lanl.gov>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

  * Neither the name of the NetworkX Developers nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
\"AS IS\" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
'''
Copied without modification from networkx 2.0 development version:
http://pelegm-networkx.readthedocs.io/en/latest/_modules/networkx/algorithms/community/centrality.html#girvan_newman
#-*- coding: utf-8 -*-  line needs to be first line
'''


import networkx as nx

__all__ = ['girvan_newman']


def girvan_newman(G, weight=None):
    '''
    Find communities in graph using Girvan–Newman method.
    
    Parameters
    ----------
    G : NetworkX graph

    weight : string, optional (default=None)
       Edge data key corresponding to the edge weight.

    Returns
    -------
    List of tuples which contains the clusters of nodes.

    Examples
    --------
    >>> G = nx.path_graph(10)
    >>> comp = girvan_newman(G)
    >>> comp[0]
    ([0, 1, 2, 3, 4], [8, 9, 5, 6, 7])

    Notes
    -----
    The Girvan–Newman algorithm detects communities by progressively removing
    edges from the original graph. Algorithm removes edge with the highest
    betweenness centrality at each step. As the graph breaks down into pieces,
    the tightly knit community structure is exposed and result can be depicted
    as a dendrogram.
    '''
    g = G.copy().to_undirected()
    components = []
    while g.number_of_edges() > 0:
        _remove_max_edge(g, weight)
        components.append(tuple(list(H)
                                for H in nx.connected_component_subgraphs(g)))
    return components



def _remove_max_edge(G, weight=None):
    """
    Removes edge with the highest value on betweenness centrality.

    Repeat this step until more connected components than the connected
    components of the original graph are detected.

    It is part of Girvan–Newman algorithm.

    :param G: NetworkX graph
    :param weight: string, optional (default=None) Edge data key corresponding
    to the edge weight.
    """
    number_components = nx.number_connected_components(G)
    while nx.number_connected_components(G) <= number_components:
        betweenness = nx.edge_betweenness_centrality(G, weight=weight)
        max_value = max(betweenness.values())
        # Use a list of edges because G is changed in the loop
        for edge in list(G.edges()):
            if betweenness[edge] == max_value:
                G.remove_edge(*edge)



from itertools import groupby
import re

def fasta_iter(fasta_name):
    '''
    Given a fasta file. yield tuples of header,  sequence
    '''
    with open(fasta_name, 'rb') as fh:
        faiter = (x[1] for x in groupby(fh,  
                                        lambda line: line[0] == ">"))
    
        for header in faiter:
            header = header.next()[1:].strip() # drop the '>'
            #join all sequence lines into one string
            seq = "".join(s.strip() for s in faiter.next()) 
            yield header,  seq

def fasta_to_dict(fasta_input_name):
    ''' Reads: Fasta-formatted sequence file.
    Returns dictionary: {fasta_id: sequence_string}
    '''
    fasta_sequences = fasta_iter(fasta_input_name)
    id_sequence_dict = {}
    for fs in fasta_sequences:
        id_sequence_dict[fs[0]]=fs[1]
    return id_sequence_dict




#! /usr/bin/env python 

from collections import defaultdict
from scipy import stats

from napa.utils.general import *
from napa.seq.BioSeq import *


class AlnMut(object):

    def __init__(self, seqs=set(), mut_str = None):
        self.mut_str = mut_str
        self.seqs = set(seqs)
    
    def add_seq(self,seq):
        self.seqs.add(seq) 
    
    def add_seqs(self,seqs):
        self.seqs.update(seqs)

    def __repr__(self):
        return self.mut_str


class AlnMutPair(object):
    '''
    Pair of muts from seq alignment.
    '''
    def __init__(self, seqs = set(), mut_pair = None):

        # Unique set of seqs containing mut
        if not len(seqs):
            raise ValueError("No seqs associated with mut pair.")
        else:
            self.seqs = set(seqs)

        if not mut_pair:
            raise ValueError("No muts given for mut pair")
        else:
            # Mutations in mut_pair should be AlnMut instances
            self.mut_pair = mut_pair 

    def __repr__(self):
        return '%s\t%s'%(str(self.mut_pair[0]), str(self.mut_pair[1])) 

    def add_seq(self, seq):
        self.seqs.update(seq) 


    def get_contingency_table(self, all_seqs):
        ''' Contingency table of occurrence counts of muts
        individually or together. 
        Occurrence count = number of seqs containing mut.
        '''

        occur01 = len(self.seqs) #occurrence count of both muts

        # Number of seqs containing first or second mut, not both 
        occur0 = len(self.mut_pair[0].seqs - self.seqs) 
        occur1 = len(self.mut_pair[1].seqs - self.seqs)

        # Number of seqs containing neither mut
        occur_other = len(all_seqs \
                        - self.mut_pair[0].seqs \
                        - self.mut_pair[1].seqs)

        self.contingency_table = [[occur01, occur0], [occur1, occur_other]]
        self.num_seqs = len(all_seqs)


    def get_jaccard_weight(self, min_co_occur = 0):
        ''' Jaccard weight with 
        1. minimum requirement for coocurrence count
        2. if no minimum is set, introduce correction for pairs 
        of muts occurring only once, and only together in one seq.'''
        [[occur01, occur0], [occur1, occur_other]] = self.contingency_table
        if min_co_occur == 0:
            self.jaccard = (float(occur01) - 1./self.num_seqs) / \
                            (float(occur01 + occur0 + occur1))            
        elif occur01 < min_co_occur:
            self.jaccard = 0.
        else:
            self.jaccard = (float(occur01) - 1./self.num_seqs) / \
                            (float(occur01 + occur0 + occur1))


    def get_fisher_pval_weight(self):
        ''' Unlike Jaccard, the smaller the p-value, 
        the greater the association '''
        self.odds_ratio_less, self.p_val_less = \
                        stats.fisher_exact(self.contingency_table, 
                                           alternative = 'less')

        self.odds_ratio_more, self.p_val_more = \
                        stats.fisher_exact(self.contingency_table, 
                                           alternative = 'greater')



class AlnMutPairSet(object):
    '''
    Set of mut pairs and associated metrics from an MSA.
    '''
    def __init__(self, aln_fasta_file = '', aln_pos = [], 
                 wt_seq_str = '', wt_id = '', pos_subset = [],
                 seqid_to_prot_func = {'':''}, 
                 seqid_to_prot_func_file = '',
                 sel_prot_func = [''], 
                 sel_prot_func_file = ''):

        # Alignment from input fasta file
        # Do not select sequences by function yet
        # (that would remove the WT/reference sequence from set
        self.aln = BioSeqAln(aln_fasta_file = aln_fasta_file,
                             aln_pos = aln_pos, 
                             pos_subset = pos_subset,
                             annot_key = 'function',
                             seqid_to_annot = seqid_to_prot_func, 
                             seqid_to_annot_file = seqid_to_prot_func_file,
                             sel_annot_key = '', 
                             sel_annot_list = [''],
                             sel_annot_file = '')

        # Get mutations from ancestral/wt sequence 
        # for each sequence in alignment
        # option to focus on subset of aln. positions
        self.get_wt_seq(wt_id, wt_seq_str)
        self.wt_seq.extract_pos(pos_subset)


        #Pick out sequences with selected function of interest
        #(Alignment may consist of sequences with different annotations)
        self.aln.subset_annot_seq_dict(sel_annot_key = 'function',
                                       sel_annot_list = sel_prot_func,
                                       sel_annot_file = sel_prot_func_file)

        self.aln.get_seq_muts(self.wt_seq)
        # stderr_write([self.aln_func.seqid_to_mut])

        # Get all possible mut pairs in alignment, 
        #given selected function
        self.get_aln_mut_pairs()
        self.print_stats()

    
    def get_wt_seq(self, wt_id, wt_seq_str):
        '''
        Obtain WT/ancestral sequence to which all other sequences in 
        the alignment should be compared. Can provide id+sequence,
        or the id of a sequence already in the alignment.
        '''
        if len(wt_id) and len(wt_seq_str) == self.aln.length:
            self.wt_seq = BioSeq(seq_id = wt_id, seq_str = wt_seq_str,
                                 seq_type = 'Protein', 
                                 seq_pos_list = self.aln_pos)
        elif len(wt_id) and len(wt_seq_str) != self.aln.length:
            if wt_id in self.aln.seqid_to_seq:
                stderr_write(['AlnMutPairSet WARNING:',
                              'Sequence length does not match alignment,',
                              'but sequence id in alignment. Replacing',
                              'with corresp. alignment sequence.'])
                self.wt_seq = self.aln.seqid_to_seq[wt_id]
            else:
                stderr_write(['Invalid  input for WT sequence.',
                              'Please enter valid id already in alignment.',
                               'or a new sequence.'])
                              
        elif not len(wt_id) and len(wt_seq_str) == self.aln.length:
            self.wt_seq = BioSeq(seq_id = 'Wild_type', 
                                 seq_str = wt_seq_str,
                                 seq_type = 'Protein', 
                                 seq_pos_list = self.aln_pos)

        else:
            stderr_write(['Invalid  input for WT sequence.',
                          'Please enter valid id already in alignment.',
                          'or a new sequence.'])
        
    def get_aln_mut_pairs(self):
        ''' Extract pairs of mutations occurring in at least one seq.'''

        self.mut_str_to_obj, self.mut_pair_to_obj = {}, {}
        for seqid in self.aln.seqid_to_mut:
            for mut_str in self.aln.seqid_to_mut[seqid]:
                if mut_str not in self.mut_str_to_obj:
                    self.mut_str_to_obj[mut_str] = \
                            AlnMut(seqs = set([seqid]), 
                                   mut_str = mut_str)
                else:
                    self.mut_str_to_obj[mut_str].add_seq(seqid)

            for mut_str_pair in list_pairs(self.aln.seqid_to_mut[seqid]):
                if mut_str_pair not in self.mut_pair_to_obj:
                    self.mut_pair_to_obj[mut_str_pair] = \
                        AlnMutPair(seqs = set([seqid]), 
                                   mut_pair = (self.mut_str_to_obj[mut_str_pair[0]],
                                               self.mut_str_to_obj[mut_str_pair[1]]))
                else:
                    self.mut_pair_to_obj[mut_str_pair].add_seq(seqid)

        # get contingency table for mutation pair stats
        for mut_pair in self.mut_pair_to_obj:
            self.mut_pair_to_obj[mut_pair].get_contingency_table(all_seqs = \
                                           set(self.aln.seqid_to_mut.keys()))
                        


    def write_jaccard_weights_table(self, min_co_occur = 0):
        out_str = '\t'.join(['_Mut1_', '_Mut2_', 'Num_sequences_co-occur', 
                             'Contingency_Tab(common_1_2_other)',
                             'Jaccard_index']) + '\n'
        

        for mut_str_pair in self.mut_pair_to_obj:
            aln_mut_pair = self.mut_pair_to_obj[mut_str_pair]
            aln_mut_pair.get_jaccard_weight(min_co_occur = min_co_occur)

            if aln_mut_pair.jaccard == 0.:
                continue

            to_print = [str(aln_mut_pair)]
            to_print.append(str(len(aln_mut_pair.seqs)))
            to_print.append(str(aln_mut_pair.contingency_table))
            to_print.append(str(aln_mut_pair.jaccard))

            out_str +=  '\t'.join(to_print) + '\n'

        return out_str
                                         
    def write_jaccard_weights_network(self, min_co_occur = 0):
        #MCR - Remove header to have network file same as the one from phylo_mut_pairs
        #out_str = '\t'.join(['Source', 'Target', 'weight']) + '\n'
        out_str = '';
        

        for mut_str_pair in self.mut_pair_to_obj:
            aln_mut_pair = self.mut_pair_to_obj[mut_str_pair]
            aln_mut_pair.get_jaccard_weight(min_co_occur = min_co_occur)

            if aln_mut_pair.jaccard == 0.:
                continue

            out_str += str(aln_mut_pair) + '\t' + str(aln_mut_pair.jaccard) + '\n'

        return out_str


    def __repr__(self):
        '''
        Writes table of mutation pairs and statistics.
        '''
        out_str = '\t'.join(['_Mut1', '_Mut2', 'numSharedSeqs', 
                             'contTab(common_1_2_other)']) + '\n'
        for aln_mut_pair in self.aln_mut_pairs:
            to_print = [str(aln_mut_pair)]
            to_print.append(str(len(aln_mut_pair.seqs)))
            to_print.append(str(aln_mut_pair.contingency_table))
            out_str +=  '\t'.join(to_print) + '\n'

        return out_str


    def print_stats(self):

        stderr_write(['Alignment depth and length:', self.aln.depth, self.aln.length])

        stderr_write(["Total positions considered:", len(self.aln.aln_pos)])
                     
        stderr_write(["Number of functional mutation pairs", 
                      len(self.mut_pair_to_obj)])





import argparse
import sys
from os import listdir
from os.path import isfile, join
import itertools
from collections import defaultdict

from napa.utils.general import *
from napa.mutpair.phylo import * # Netw. construct from alignment


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('-a', dest = 'alnFASTA', required = True, 
                        help = 'Path of multiple sequence alignment fasta file.')

    parser.add_argument('-pl', dest = 'posList', required = True, 
                        help = 'File containing protein residue positions ' + \
                        'corresponding to each column in the alignment.')

    parser.add_argument('-d', dest = 'treeDistThresh', type = float, required = True, 
                        help = 'Maximum distance between mutations on the tree. ' + \
                        'Mutations further appart not linked in network.')


    parser.add_argument('-wt', dest = 'wtId', required = True,  
                        help = 'Id of wildtype or ancestral sequence, ' + \
                        'assuming it is already in alignment. Can also provide '+ \
                        'separate FASTA with wildtype sequence.')

    parser.add_argument('-ps', dest = 'posSubset', required = True, 
                        help = 'File containing subset of aln/protein positions' + \
                        'to be considered in network')

    parser.add_argument('-af', dest = 'protFunc', required = True, 
                        help = 'File containing known protein functional annot. ' + \
                        'of sequences in alignment.')

    parser.add_argument('-ft', dest = 'funcTransitions', required = True, 
                        help = 'String/File containing allowed transitions between ' + \
                        'protein functions in the evolution of a final function.')

    parser.add_argument('-sf', dest = 'selProtFuncs', required = True, 
                        help = 'Path of file containing the selected final function.')


    parser.add_argument('-tl', dest = 'treeFileList', required = True,
                        help = 'File with a list of tree files from the ensemble.')
        
    parser.add_argument('-is', dest = 'intSeqFileList', required = True,
                        help = 'File with a list of fasta files where the internal' + \
                        'node sequences are stored.')
        
    parser.add_argument('-if', dest = 'intFuncFileList', required = True,
                        help = 'File with a list of fasta files where the internal' + \
                        'node functions are stored.')

    parser.add_argument('-mt', dest = 'mutPairType', type = str, required = True,
                        help = 'The type of network: dir(ected) or undir(ected).')
    
    parser.add_argument('-pt', dest = 'pvalThresh', type = float, required = True,
                        help = 'The maximum p-value for Fisher\'s exact test.' + \
                        'Used to extract significantly associated mut. pairs.')
    

    parser.add_argument('-nf', dest = 'netFile', required = True, 
                        help = 'Outputfile for result network.')

    args = parser.parse_args()

    return args


class PhyloNetInput(object):
    '''
    Prepare input files for alignment-based network reconstruction.
    '''
    def __init__(self, args):

        # List of positions/column numbers in alignment
        # If not provided, positions start with 1,2..alignment_length
        self.pos_list = [int(rec) for rec in parse_column(args.posList)]

        # List of protein residue positions considered for network.
        # (remove residues from signaling peptide, or other domains not studied).
        self.pos_subset = [int(rec) for rec in parse_column(args.posSubset)]

        # Functional transitions between phylogeny sequences allowed
        # (evolution of specific function)
        self.prot_func_transitions = parse_keyval_dict(args.funcTransitions)

        # List of absolute paths for phylogenetic trees in ensemble
        self.tree_files = parse_column(args.treeFileList)

        # List of internal node sequence fasta files for each tree
        self.int_seq_files = parse_column(args.intSeqFileList)

        # List of internal node function files for each tree
        self.int_func_files = parse_column(args.intFuncFileList)

        # check we have required inputs for each tree
        if len(self.tree_files) != len(self.int_seq_files):
            stderr_write(['Unequal number of tree files and ',
                          'internal node files.',
                          'Using minimum number of files of both.'])

            min_trees = min(len(self.tree_files), len(self.int_seq_files))
            self.tree_files = self.tree_files[0:min_trees]
            self.int_seq_files = self.int_seq_files[0:min_trees]


def main():
    args = parse_args()
    inp = PhyloNetInput(args)
    for i, (tf, isf, iff) in enumerate(zip(inp.tree_files, 
                                           inp.int_seq_files, inp.int_func_files)):

        stderr_write(['\n','Extracting mutation pairs for tree file:\n',tf])
        # Set of mutation pairs occurring along a single tree
        phylo_mut_pair_set = PhyloMutPairSet(leaf_fasta_file = args.alnFASTA,
                                             int_node_fasta_file = isf,
                                             aln_pos = inp.pos_list,
                                             wt_id = args.wtId,
                                             pos_subset = inp.pos_subset,
                                             leaf_seqid_to_prot_func_file = args.protFunc,
                                             int_seqid_to_prot_func_file = iff,
                                             prot_func_transitions_file = args.funcTransitions,
                                             sel_prot_func_file = args.selProtFuncs,
                                             dist_thresh = args.treeDistThresh,
                                             tree_nwk_file = tf,
                                             mut_pair_type = args.mutPairType)
        # Combined sets of mutations from all trees 
        # The number of trees in ensemble with a significant right-tailed p-value
        # are the weights output for each mutation pair
        if i == 0:
            cmb_mut_pair_set = CombinedPhyloMutPairs(phylo_mut_pair_set, args.pvalThresh)
        else:
            cmb_mut_pair_set.add_mut_pairs(phylo_mut_pair_set)

        
    with open(args.netFile, 'wb') as f:
        f.write(str(cmb_mut_pair_set))

if __name__ == '__main__': main()        



from setuptools import setup;
from setuptools import find_packages;

setup(name='napa',
      version='0.86',
      description='Network Analysis of Protein Adaptation (intra-protein residue coevolution network construction and analysis)',
      url='http://karchinlab.org/napa',
      author='Violeta Beleva-Guthrie',
      author_email='vbeleva@gmail.com',
      license='',
      packages=find_packages(exclude=['tests']),
      install_requires=[
          'numpy',
          'scipy',
          'networkx',
          'ete2'
      ],
      classifiers=[
      'Development Status :: 4 - Beta',
      'Intended Audience :: Developers',
      'Topic :: Scientific/Engineering :: Bio-Informatics'
      ],
      zip_safe=True)